import pigpio
import RPi.GPIO as GPIO
import time

# Hier definieren wir die Pins die wir benutzen wollen
PWM_PIN = 18  # Der Pin für den Servo-Motor
ULTRASCHALL_TRIGGER_PIN = 23  # Der Pin für den Trigger des Ultraschallsensors
ULTRASCHALL_ECHO_PIN = 24  # Der Pin für das Echo des Ultraschallsensors
IR_SENSOR_PIN = 25  # Der Pin für den IR-Sensor

# Schranken-Einstellungen für den Servo-Motor
FREQUENCY = 50  # Die Frequenz des PWM Signals für den Servo
MIN_PULSE_WIDTH = 500  # Der minimale Pulswert für den Servo
MAX_PULSE_WIDTH = 2500  # Der maximale Pulswert für den Servo

# Zustands-Flags für die Schranke und das Fahrzeug
schranke_oben = False  # Ob die Schranke oben ist oder nicht
fahrzeug_erkannt = False  # Ob ein Fahrzeug erkannt wurde
fahrzeug_flag = False  # Flag für den IR-Sensor, wird im Interrupt gesetzt
letzte_erkennung = 0  
debounce_zeit = 0.2  # 200 ms

# Diese Funktion berechnet den Pulswert für den Servo abhängig vom Winkel
def angle_to_pulse_width(angle):
    return int(MIN_PULSE_WIDTH + (angle / 180.0) * (MAX_PULSE_WIDTH - MIN_PULSE_WIDTH))

# Diese Funktion gibt die gemessene Distanz des Ultraschallsensors zurück
def get_ultrasound_distance():
    GPIO.output(ULTRASCHALL_TRIGGER_PIN, GPIO.HIGH)  # Trigger-Pin für den Ultraschallsensor setzen
    time.sleep(0.00001)  # 10 Mikrosekunden warten
    GPIO.output(ULTRASCHALL_TRIGGER_PIN, GPIO.LOW)  # Trigger-Pin wieder ausschalten

    # Initialisierung der Zeitvariablen bevor die Messung beginnt
    start_time = time.time()
    stop_time = time.time()

    # Startzeit setzen wenn Echo-Pin auf HIGH geht
    while GPIO.input(ULTRASCHALL_ECHO_PIN) == 0:
        start_time = time.time()

    # Stopzeit setzen wenn Echo-Pin auf LOW geht
    while GPIO.input(ULTRASCHALL_ECHO_PIN) == 1:
        stop_time = time.time()

    # Berechnung der Entfernung basierend auf der gemessenen Zeit
    time_elapsed = stop_time - start_time
    return (time_elapsed * 34300) / 2  # Schallgeschwindigkeit in cm/s

# Diese Funktion verarbeitet das Signal des IR-Sensors außerhalb des Interrupts
def handle_ir_sensor():
    global fahrzeug_erkannt, schranke_oben, fahrzeug_flag, letzte_erkennung

    if fahrzeug_flag:  # Wurde durch den Interrupt gesetzt
        aktuelle_zeit = time.time()
        
        # Debounce-Prüfung: Nur weitermachen, wenn 200 ms vergangen sind
        if (aktuelle_zeit - letzte_erkennung) > debounce_zeit:
            fahrzeug_flag = False  # Flag zurücksetzen
            letzte_erkennung = aktuelle_zeit  # Zeitstempel aktualisieren
            
            if not fahrzeug_erkannt:
                fahrzeug_erkannt = True  # Fahrzeug erkannt
                print("Fahrzeug erkannt!")  # Status ausgeben
            else:
                # In einer Schleife prüfen, ob der Sensor wirklich wieder frei ist
                while pi.read(IR_SENSOR_PIN) == 0:
                    time.sleep(0.1)

                # Wenn der Sensor frei bleibt, wird die Schranke geschlossen
                fahrzeug_erkannt = False
                print("Fahrzeug hat Bereich verlassen.")  # Status ausgeben
                if schranke_oben:
                    schranke_oben = False
                    pi.set_servo_pulsewidth(PWM_PIN, angle_to_pulse_width(0))  # Schranke schließen
                    print("Schranke geschlossen.")

# Minimierter Interrupt-Handler
def ir_callback(gpio, level, tick):
    global fahrzeug_flag
    fahrzeug_flag = True  # FLAGGGG (Interrupt so kurz wie möglich halten :D)

try:
    # Initialisiere die pigpio Bibliothek
    pi = pigpio.pi()
    if not pi.connected:
        exit()

    # Set up die GPIO Pins mit der RPi.GPIO Bibliothek
    GPIO.setmode(GPIO.BCM)
    GPIO.setup(ULTRASCHALL_TRIGGER_PIN, GPIO.OUT)
    GPIO.setup(ULTRASCHALL_ECHO_PIN, GPIO.IN)
    GPIO.setup(PWM_PIN, GPIO.OUT)
    GPIO.output(ULTRASCHALL_TRIGGER_PIN, False)

    # Servo initialisieren und anhalten
    GPIO.output(PWM_PIN, GPIO.LOW)

    # Setzet den IR-Sensor auf INPUT und wartet auf einen Interrupt
    pi.set_mode(IR_SENSOR_PIN, pigpio.INPUT)
    pi.callback(IR_SENSOR_PIN, pigpio.EITHER_EDGE, ir_callback)

    # Die Haupt-Polling-Schleife
    while True:
        handle_ir_sensor()  # Verarbeitung des IR-Signals außerhalb des Interrupts

        # Holt die gemessene Entfernung vom Ultraschallsensor
        distance = get_ultrasound_distance()
        print(f"Gemessene Entfernung: {distance:.2f} cm")  # Entfernung ausgeben

        # Wenn die Entfernung unter 20 cm liegt und die Schranke noch nicht oben ist
        if distance < 20 and not schranke_oben:
            schranke_oben = True
            pi.set_servo_pulsewidth(PWM_PIN, angle_to_pulse_width(90))  # Schranke hochfahren
            print("Schranke hochgefahren.")

        time.sleep(0.5)

except KeyboardInterrupt:
    pass

finally:
    pi.set_servo_pulsewidth(PWM_PIN, 0)
    GPIO.cleanup()
    pi.stop()
